@@ -22,7 +22,6 @@ use std::path::{Path, PathBuf};
 use std::process::{Command, Output};
 use std::str;
 use tempfile::{tempdir, TempDir};
-use walkdir::WalkDir;
 use zip::{self, ZipWriter};
 
 /// Allows writing the module to a wheel or add it directly to the virtualenv
@@ -81,16 +80,13 @@ pub struct PathWriter {
 }
 
 impl PathWriter {
-    /// Creates a [ModuleWriter] that adds the modul to the current virtualenv
+    /// Creates a [ModuleWriter] that adds the module to the current virtualenv
     pub fn venv(target: &Target, venv_dir: &Path, bridge: &BridgeModel) -> Result<Self> {
-        let interpreter = PythonInterpreter::check_executable(
-            target.get_venv_python(&venv_dir),
-            &target,
-            &bridge,
-        )?
-        .ok_or_else(|| {
-            anyhow!("Expected `python` to be a python interpreter inside a virtualenv ಠ_ಠ")
-        })?;
+        let interpreter =
+            PythonInterpreter::check_executable(target.get_venv_python(&venv_dir), target, bridge)?
+                .ok_or_else(|| {
+                    anyhow!("Expected `python` to be a python interpreter inside a virtualenv ಠ_ಠ")
+                })?;
 
         let base_path = target.get_venv_site_package(venv_dir, &interpreter);
 
@@ -231,7 +227,7 @@ impl ModuleWriter for WheelWriter {
             .unix_permissions(permissions)
             .compression_method(compression_method);
         self.zip.start_file(target.clone(), options)?;
-        self.zip.write_all(&bytes)?;
+        self.zip.write_all(bytes)?;
 
         let hash = base64::encode_config(&Sha256::digest(bytes), base64::URL_SAFE_NO_PAD);
         self.record.push((target, hash, bytes.len()));
@@ -266,11 +262,34 @@ impl WheelWriter {
             wheel_path,
         };
 
-        write_dist_info(&mut builder, &metadata21, &tags)?;
+        write_dist_info(&mut builder, metadata21, tags)?;
 
         Ok(builder)
     }
 
+    /// Add a pth file to wheel root for editable installs
+    pub fn add_pth(
+        &mut self,
+        project_layout: &ProjectLayout,
+        metadata21: &Metadata21,
+    ) -> Result<()> {
+        match project_layout {
+            ProjectLayout::Mixed {
+                ref python_module, ..
+            } => {
+                let absolute_path = python_module.canonicalize()?;
+                if let Some(python_path) = absolute_path.parent().and_then(|p| p.to_str()) {
+                    let name = metadata21.get_distribution_escaped();
+                    self.add_bytes(format!("{}.pth", name), python_path.as_bytes())?;
+                } else {
+                    println!("⚠️ source code path contains non-Unicode sequences, editable installs may not work.");
+                }
+            }
+            ProjectLayout::PureRust { .. } => {}
+        }
+        Ok(())
+    }
+
     /// Creates the record file and finishes the zip
     pub fn finish(mut self) -> Result<PathBuf, io::Error> {
         let compression_method = if cfg!(feature = "faster-tests") {
@@ -469,7 +488,7 @@ fn cffi_header(crate_dir: &Path, tempdir: &TempDir) -> Result<PathBuf> {
 /// in `__init__.py` to load the shared library into a module called `lib`.
 pub fn generate_cffi_declarations(crate_dir: &Path, python: &Path) -> Result<String> {
     let tempdir = tempdir()?;
-    let header = cffi_header(&crate_dir, &tempdir)?;
+    let header = cffi_header(crate_dir, &tempdir)?;
 
     let ffi_py = tempdir.as_ref().join("ffi.py");
 
@@ -490,7 +509,7 @@ recompiler.make_py_source(ffi, "ffi", r"{ffi_py}")
         header = header.display(),
     );
 
-    let output = call_python(&python, &["-c", &cffi_invocation])?;
+    let output = call_python(python, &["-c", &cffi_invocation])?;
     let install_cffi = if !output.status.success() {
         // First, check whether the error was cffi not being installed
         let last_line = str::from_utf8(&output.stderr)?.lines().last().unwrap_or("");
@@ -499,7 +518,7 @@ recompiler.make_py_source(ffi, "ffi", r"{ffi_py}")
             // We don't want to modify any global environment
             // https://stackoverflow.com/a/42580137/3549270
             let output = call_python(
-                &python,
+                python,
                 &["-c", "import sys\nprint(sys.base_prefix != sys.prefix)"],
             )?;
 
@@ -508,7 +527,7 @@ recompiler.make_py_source(ffi, "ffi", r"{ffi_py}")
                 "False" => false,
                 _ => {
                     println!(
-                        "⚠  Failed to determine whether python at {:?} is running inside a virtualenv",
+                        "⚠️ Failed to determine whether python at {:?} is running inside a virtualenv",
                         &python
                     );
                     false
@@ -523,13 +542,13 @@ recompiler.make_py_source(ffi, "ffi", r"{ffi_py}")
 
     // If there was success or an error that was not missing cffi, return here
     if !install_cffi {
-        return handle_cffi_call_result(&python, tempdir, &ffi_py, &output);
+        return handle_cffi_call_result(python, tempdir, &ffi_py, &output);
     }
 
-    println!("⚠  cffi not found. Trying to install it");
+    println!("⚠️ cffi not found. Trying to install it");
     // Call pip through python to don't do the wrong thing when python and pip
     // are coming from different environments
-    let output = call_python(&python, &["-m", "pip", "install", "cffi"])?;
+    let output = call_python(python, &["-m", "pip", "install", "cffi"])?;
     if !output.status.success() {
         bail!(
             "Installing cffi with `{:?} -m pip install cffi` failed: {}\n--- Stdout:\n{}\n--- Stderr:\n{}\n---\nPlease install cffi yourself.",
@@ -542,8 +561,8 @@ recompiler.make_py_source(ffi, "ffi", r"{ffi_py}")
     println!("🎁 Installed cffi");
 
     // Try again
-    let output = call_python(&python, &["-c", &cffi_invocation])?;
-    handle_cffi_call_result(&python, tempdir, &ffi_py, &output)
+    let output = call_python(python, &["-c", &cffi_invocation])?;
+    handle_cffi_call_result(python, tempdir, &ffi_py, &output)
 }
 
 /// Extracted into a function because this is needed twice
@@ -572,6 +591,7 @@ fn handle_cffi_call_result(
 }
 
 /// Copies the shared library into the module, which is the only extra file needed with bindings
+#[allow(clippy::too_many_arguments)]
 pub fn write_bindings_module(
     writer: &mut impl ModuleWriter,
     project_layout: &ProjectLayout,
@@ -579,11 +599,11 @@ pub fn write_bindings_module(
     artifact: &Path,
     python_interpreter: Option<&PythonInterpreter>,
     target: &Target,
-    develop: bool,
+    editable: bool,
 ) -> Result<()> {
     let ext_name = project_layout.extension_name();
     let so_filename = match python_interpreter {
-        Some(python_interpreter) => python_interpreter.get_library_name(&ext_name),
+        Some(python_interpreter) => python_interpreter.get_library_name(ext_name),
         // abi3
         None => {
             if target.is_unix() {
@@ -601,10 +621,12 @@ pub fn write_bindings_module(
             ref rust_module,
             ..
         } => {
-            write_python_part(writer, python_module, &module_name)
-                .context("Failed to add the python module to the package")?;
+            if !editable {
+                write_python_part(writer, python_module, &module_name)
+                    .context("Failed to add the python module to the package")?;
+            }
 
-            if develop {
+            if editable {
                 let target = rust_module.join(&so_filename);
                 fs::copy(&artifact, &target).context(format!(
                     "Failed to copy {} to {}",
@@ -613,8 +635,10 @@ pub fn write_bindings_module(
                 ))?;
             }
 
-            let relative = rust_module.strip_prefix(python_module.parent().unwrap())?;
-            writer.add_file_with_permissions(relative.join(&so_filename), &artifact, 0o755)?;
+            if !editable {
+                let relative = rust_module.strip_prefix(python_module.parent().unwrap())?;
+                writer.add_file_with_permissions(relative.join(&so_filename), &artifact, 0o755)?;
+            }
         }
         ProjectLayout::PureRust {
             ref rust_module, ..
@@ -624,7 +648,11 @@ pub fn write_bindings_module(
             // Reexport the shared library as if it were the top level module
             writer.add_bytes(
                 &module.join("__init__.py"),
-                format!("from .{} import *\n", module_name).as_bytes(),
+                format!(
+                    "from .{module_name} import *\n\n__doc__ = {module_name}.__doc__\n",
+                    module_name = module_name
+                )
+                .as_bytes(),
             )?;
             let type_stub = rust_module.join(format!("{}.pyi", module_name));
             if type_stub.exists() {
@@ -643,6 +671,7 @@ pub fn write_bindings_module(
 }
 
 /// Creates the cffi module with the shared library, the cffi declarations and the cffi loader
+#[allow(clippy::too_many_arguments)]
 pub fn write_cffi_module(
     writer: &mut impl ModuleWriter,
     project_layout: &ProjectLayout,
@@ -650,9 +679,9 @@ pub fn write_cffi_module(
     module_name: &str,
     artifact: &Path,
     python: &Path,
-    develop: bool,
+    editable: bool,
 ) -> Result<()> {
-    let cffi_declarations = generate_cffi_declarations(&crate_dir, python)?;
+    let cffi_declarations = generate_cffi_declarations(crate_dir, python)?;
 
     let module;
 
@@ -662,10 +691,12 @@ pub fn write_cffi_module(
             ref rust_module,
             ref extension_name,
         } => {
-            write_python_part(writer, python_module, &module_name)
-                .context("Failed to add the python module to the package")?;
+            if !editable {
+                write_python_part(writer, python_module, &module_name)
+                    .context("Failed to add the python module to the package")?;
+            }
 
-            if develop {
+            if editable {
                 let base_path = python_module.join(&module_name);
                 fs::create_dir_all(&base_path)?;
                 let target = base_path.join("native.so");
@@ -681,7 +712,9 @@ pub fn write_cffi_module(
 
             let relative = rust_module.strip_prefix(python_module.parent().unwrap())?;
             module = relative.join(extension_name);
-            writer.add_directory(&module)?;
+            if !editable {
+                writer.add_directory(&module)?;
+            }
         }
         ProjectLayout::PureRust {
             ref rust_module, ..
@@ -700,9 +733,11 @@ pub fn write_cffi_module(
         }
     };
 
-    writer.add_bytes(&module.join("__init__.py"), cffi_init_file().as_bytes())?;
-    writer.add_bytes(&module.join("ffi.py"), cffi_declarations.as_bytes())?;
-    writer.add_file_with_permissions(&module.join("native.so"), &artifact, 0o755)?;
+    if !editable || matches!(project_layout, ProjectLayout::PureRust { .. }) {
+        writer.add_bytes(&module.join("__init__.py"), cffi_init_file().as_bytes())?;
+        writer.add_bytes(&module.join("ffi.py"), cffi_declarations.as_bytes())?;
+        writer.add_file_with_permissions(&module.join("native.so"), &artifact, 0o755)?;
+    }
 
     Ok(())
 }
@@ -735,7 +770,9 @@ pub fn write_python_part(
     python_module: impl AsRef<Path>,
     module_name: impl AsRef<Path>,
 ) -> Result<()> {
-    for absolute in WalkDir::new(&python_module) {
+    use ignore::WalkBuilder;
+
+    for absolute in WalkBuilder::new(&python_module).hidden(false).build() {
         let absolute = absolute?.into_path();
 
         let relative = absolute.strip_prefix(python_module.as_ref().parent().unwrap())?;
